from datetime import timedelta

from basxbread.utils.celery import RepeatedTask
from celery import shared_task
from django.apps import AppConfig
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _


class CoreConfig(AppConfig):
    name = "basxconnect.core"
    default_auto_field = "django.db.models.BigAutoField"
    verbose_name = _("BasxConnect Core")

    def ready(self):
        from django.db.models.signals import post_save

        from .models import Address, Phone, Vocabulary

        shared_task(base=RepeatedTask, run_every=timedelta(hours=6))(update_addresses)

        def saveperson(sender, instance, **kwargs):
            instance.person.save()

        post_save.connect(saveperson, Phone, dispatch_uid="save_person_phone")
        post_save.connect(saveperson, Address, dispatch_uid="save_person_address")

        pre_installed_vocabulary = {
            "tag": _("Tags"),
            "title": _("Titles"),
            "salutation": _("Salutations"),
            "form_of_address": _("Forms Of Address"),
            "gender": _("Genders"),
            "naturaltype": _("Natural Person Types"),
            "legaltype": _("Legal Person Types"),
            "associationtype": _("Association Person Types"),
            "addressstatus": _("Address statuses"),
            "emailtype": _("Email Types"),
            "urltype": _("URL Types"),
            "phonetype": _("Phone Types"),
            "addresstype": _("Address Types"),
        }
        try:
            for slug, name in pre_installed_vocabulary.items():
                Vocabulary.objects.get_or_create(slug=slug, defaults={"name": name})
        except Exception:  # nosec this is trivial, no service interuption
            pass


def update_addresses():
    from .models import Postal

    for address in Postal.objects.filter(valid_until__lt=now().date()):
        address.person.save()



# Generated by Django 4.1.7 on 2023-04-03 06:06

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0035_alter_email_type_alter_fax_type_and_more"),
        ("basxconnect_projects", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="project",
            name="client",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="projects",
                to="core.person",
                verbose_name="Client",
            ),
        ),
    ]



import hashlib
from typing import List

import mailchimp_marketing
from django.conf import settings
from dynamic_preferences.registries import global_preferences_registry
from mailchimp_marketing.api_client import ApiClientError

from basxconnect.mailer_integration.abstract import mailer
from basxconnect.mailer_integration.abstract.mailer import MailerPerson, MailingInterest
from basxconnect.mailer_integration.mailchimp.parsing import (
    create_mailer_person_from_raw,
)
from basxconnect.mailer_integration.models import Interest


class Mailchimp(mailer.AbstractMailer):
    def _dynamic_setting(self, name: str):
        return global_preferences_registry.manager()[name]

    def _client(self):
        client = mailchimp_marketing.Client()
        client.set_config(
            {
                "api_key": self._dynamic_setting("mailchimp__api_key"),
                "server": self._dynamic_setting("mailchimp__server"),
            }
        )
        return client

    def name(self) -> str:
        return "Mailchimp"

    def get_person_count(self) -> int:
        return self._get_segment(1, 0)["total_items"]

    def get_persons(self, count: int, offset: int) -> List[MailerPerson]:
        segment = self._get_segment(count, offset)
        return [
            create_mailer_person_from_raw(raw_person)
            for raw_person in segment["members"]
        ]

    def _get_segment(self, count, offset):
        return self._client().lists.get_segment_members_list(
            list_id=self._dynamic_setting("mailchimp__list_id"),
            segment_id=self._dynamic_setting("mailchimp__segment_id"),
            count=count,
            offset=offset,
            include_cleaned=True,
            include_transactional=True,
            include_unsubscribed=True,
            # TODO: Find out why this does not work. Would be nice to request as little data as possible.
            # fields=[
            #     "email_address",
            #     "merge_fields.FNAME",
            #     "merge_fields.LNAME",
            # ],
        )

    def delete_person(self, email: str):
        self._client().lists.delete_list_member(
            self._dynamic_setting("mailchimp__list_id"),
            compute_email_hash(email),
        )

    def put_person(self, person: MailerPerson, **kwargs):
        email_hash = compute_email_hash(person.email)
        self._client().lists.set_list_member(
            self._dynamic_setting("mailchimp__list_id"),
            email_hash,
            {
                "email_address": person.email,
                "status_if_new": person.status,
                "interests": compute_interests_dict(person),
                "merge_fields": {
                    "FNAME": person.first_name or person.display_name,
                    "LNAME": person.last_name,
                    **(
                        settings.MAILCHIMP_ADDITIONAL_MERGE_FIELDS(person)
                        if hasattr(settings, "MAILCHIMP_ADDITIONAL_MERGE_FIELDS")
                        else {}
                    ),
                },
                "language": person.language,
                **kwargs,
            },
        )
        self._client().lists.update_list_member_tags(
            self._dynamic_setting("mailchimp__list_id"),
            email_hash,
            {
                "tags": [
                    {
                        "name": self._dynamic_setting("mailchimp__tag"),
                        "status": "active",
                    }
                ]
            },
        )

    def add_person(self, person: MailerPerson):
        self.put_person(person, status=person.status)

    def email_exists(self, email: str) -> bool:
        try:
            self._client().lists.get_list_member(
                self._dynamic_setting("mailchimp__list_id"),
                compute_email_hash(email),
            )
        except ApiClientError as e:
            if e.status_code == 404:
                return False
            raise e
        return True

    def change_email_address(self, old_email: str, new_email: str):
        self._client().lists.update_list_member(
            self._dynamic_setting("mailchimp__list_id"),
            compute_email_hash(old_email),
            {"email_address": new_email},
        )

    def get_interests(self):
        interests_category_id = self._dynamic_setting(
            "mailchimp__interests_category_id"
        )
        if interests_category_id:
            return [
                MailingInterest(interest["id"], interest["name"])
                for interest in self._client().lists.list_interest_category_interests(
                    self._dynamic_setting("mailchimp__list_id"),
                    interests_category_id,
                )["interests"]
            ]
        else:
            return []

    def tag(self) -> str:
        return self._dynamic_setting("mailchimp__basxconnect_tag")


def compute_interests_dict(person) -> dict:
    interests = dict(
        [(interest.external_id, False) for interest in Interest.objects.all()]
    )
    for interest_id in person.interests_ids:
        interests[interest_id] = True
    return interests


def compute_email_hash(email):
    # required by mailchimp, ignore security check
    return hashlib.md5(str(email).lower().encode()).hexdigest()  # nosec



import htmlgenerator as hg
from basxbread import layout, utils
from basxbread.contrib.document_templates.models import DocumentTemplate
from django import forms, http
from django.contrib import messages
from django.core.mail import EmailMessage
from django.shortcuts import get_object_or_404
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry

from .models import Invoice


class EmailEntry(forms.Form):
    recipient = forms.EmailField()


EmailsFormset = forms.formset_factory(EmailEntry, can_delete=True, extra=0)


class SendForm(forms.Form):
    subject = forms.CharField()
    body = forms.CharField(widget=forms.Textarea)
    template = forms.ModelChoiceField(
        queryset=DocumentTemplate.objects.filter(
            model__app_label="basxconnect_invoicing", model__model="invoice"
        )
    )


def send_message(
    request,
    pk: int,
    subject_template,
    body_template,
    attachment_template,
    is_receipt=False,
):
    invoice = get_object_or_404(Invoice, id=pk)
    global_preferences = global_preferences_registry.manager()

    our_company_name = global_preferences["general__organizationname"]
    preferred_email_type = global_preferences["invoicing__default_invoice_email_type"]

    emails = invoice.client.core_email_list.all()
    if preferred_email_type is not None:
        emails = emails.filter(type=preferred_email_type)
    else:
        emails = emails[:1]
    initial_emails = (
        [{"recipient": email.email} for email in emails]
        if emails.exists()
        else [{"recipient": ""}]
    )
    subject = utils.jinja_render(
        subject_template, invoice=invoice, our_company_name=our_company_name
    )
    body = utils.jinja_render(
        body_template, invoice=invoice, our_company_name=our_company_name
    )

    send_form = SendForm(
        initial={
            "subject": subject,
            "body": body,
            "template": attachment_template,
        }
    )
    recipient_formset = EmailsFormset(initial=initial_emails)
    if request.method == "POST":
        send_form = SendForm(
            request.POST,
            initial={
                "subject": subject,
                "body": body,
                "template": attachment_template,
            },
        )
        recipient_formset = EmailsFormset(request.POST, initial=initial_emails)
        if send_form.is_valid() and recipient_formset.is_valid():
            email = EmailMessage(
                subject=send_form.cleaned_data["subject"],
                body=send_form.cleaned_data["body"],
                to=[recp["recipient"] for recp in recipient_formset.cleaned_data],
            )
            filename, data = send_form.cleaned_data["template"].generate_document_pdf(
                invoice
            )
            email.attach(filename, data, "application/pdf")
            email.send()
            if is_receipt:
                invoice.receipt_sent = now().date()
            else:
                invoice.invoice_sent = now().date()
            invoice.save()
            messages.success(request, _("Email successfully sent"))
            return http.HttpResponse("Ok")

    return hg.DIV(
        layout.forms.Form(
            send_form,
            hg.DIV(
                layout.forms.helpers.Label(_("Recepients")),
                layout.forms.Formset.as_plain(
                    recipient_formset,
                    layout.forms.FormField("recipient", no_wrapper=True, no_label=True),
                    add_label=_("Add recipient"),
                ),
                style="max-width: 20rem",
            ),
            hg.DIV(
                layout.forms.FormField("subject", style="width: 100%"),
                layout.forms.FormField("body", style="width: 100%"),
                layout.forms.FormField("template"),
            ),
        )
    )


@utils.aslayout
def send_invoice(request, pk: int):
    global_preferences = global_preferences_registry.manager()
    return send_message(
        request,
        pk,
        global_preferences["invoicing__invoice_message_subject_template"],
        global_preferences["invoicing__invoice_message_body_template"],
        global_preferences["invoicing__default_invoice_template"],
    )


@utils.aslayout
def send_receipt(request, pk: int):
    global_preferences = global_preferences_registry.manager()
    return send_message(
        request,
        pk,
        global_preferences["invoicing__receipt_message_subject_template"],
        global_preferences["invoicing__receipt_message_body_template"],
        global_preferences["invoicing__default_receipt_template"],
        is_receipt=True,
    )



# Generated by Django 4.0.6 on 2022-11-25 08:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0033_historicalnaturalperson_maiden_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="term",
            name="disabled",
            field=models.BooleanField(
                default=False,
                help_text="Do not allow this term to be selected",
                verbose_name="Disabled",
            ),
        ),
    ]



# Generated by Django 3.2.8 on 2021-12-10 09:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mailer_integration", "0018_alter_synchronizationperson_sync_status"),
    ]

    operations = [
        migrations.AlterField(
            model_name="synchronizationresult",
            name="sync_completed_datetime",
            field=models.DateTimeField(null=True, verbose_name="Date and time"),
        ),
        migrations.AlterField(
            model_name="synchronizationresult",
            name="total_synchronized_persons",
            field=models.IntegerField(
                null=True, verbose_name="Number of persons in the mailer segment"
            ),
        ),
    ]



from basxbread import menu, utils
from django.utils.translation import gettext_lazy as _

from . import models

projectsGroup = menu.Group(_("Projects"))

urlpatterns = [
    *utils.default_model_paths(models.Project),
]

projectGroup = menu.Group(_("Projects"))
menu.registeritem(
    menu.Item(
        utils.Link(
            utils.ModelHref(models.Project, "browse"),
            models.Project._meta.verbose_name,
        ),
        projectGroup,
    )
)



import htmlgenerator as hg
from basxbread import layout
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext_lazy as _

from basxconnect.core.layouts.editperson.common import utils

R = layout.grid.Row
C = layout.grid.Col


def documenttemplates_tab():
    from basxbread.contrib.document_templates.models import DocumentTemplate

    def prefixes(context):
        templates = DocumentTemplate.objects.filter(
            model=ContentType.objects.get_for_model(context["object"])
        )
        ret = {}
        for template in templates:
            prefix = hg.mark_safe("<wbr/>")
            if " - " in template.name:
                prefix = template.name.split("-", 1)[0].strip()
            if prefix not in ret:
                ret[prefix] = []
            ret[prefix].append(template)
        return ret

    return layout.tabs.Tab(
        _("Series letters"),
        utils.grid_inside_tab(
            R(
                utils.tiling_col(
                    hg.DIV(
                        hg.Iterator(
                            hg.F(prefixes).items(),
                            "group",
                            hg.DIV(
                                hg.H3(hg.C("group.0")),
                                hg.UL(
                                    hg.Iterator(
                                        hg.C("group.1"),
                                        "template",
                                        hg.LI(
                                            layout.components.button.Button(
                                                icon="download",
                                                buttontype="tertiary",
                                            ).as_href(
                                                href=hg.F(
                                                    lambda c: c[
                                                        "template"
                                                    ].generate_document_url(c["object"])
                                                )
                                            ),
                                            hg.DIV(
                                                hg.C("template"),
                                                style="align-self: center; margin-left: 1rem",
                                            ),
                                            style="display: flex; margin-bottom: 1rem",
                                        ),
                                    )
                                ),
                            ),
                        ),
                        style="display: grid; grid-template-columns: 1fr 1fr",
                    )
                )
            )
        ),
    )



# Generated by Django 3.1.7 on 2021-05-22 11:17

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0008_auto_20210323_1456"),
    ]

    operations = [
        migrations.AlterField(
            model_name="relationship",
            name="person_a",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="relationships_to",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="relationship",
            name="person_b",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="relationships_from",
                to="core.person",
            ),
        ),
    ]



from basxbread.contrib.languages.languages import LANGUAGES
from django.conf import settings

DEFAULTS = {"PREFERRED_LANGUAGES": ("en",)}

preferred_codes = getattr(settings, "BASXCONNECT", {}).get(
    "PREFERRED_LANGUAGES", DEFAULTS["PREFERRED_LANGUAGES"]
)
PREFERRED_LANGUAGES = [i for i in LANGUAGES if i[0] in preferred_codes]
PREFERRED_LANGUAGES.sort(key=lambda e: preferred_codes.index(e[0]))

MIN_CHARACTERS_DYNAMIC_SEARCH = getattr(settings, "MIN_CHARACTERS_DYNAMIC_SEARCH", 3)



import logging
from typing import List

from django.conf import settings

from basxconnect.mailer_integration.abstract.mailer import MailerPerson

logger = logging.getLogger(__name__)


def create_mailer_person_from_raw(person: dict) -> MailerPerson:
    return MailerPerson(
        first_name=first_name(person),
        last_name=last_name(person),
        display_name=display_name(person),
        email=email(person),
        interests_ids=interests_ids(person),
        status=status(person),
        country=country(person),
        postcode=postcode(person),
        address=address(person),
        city=city(person),
        language=language(person),
        persontype=persontype(person),
    )


def city(raw_person):
    custom_reader = getattr(settings, "MAILCHIMP_CITY_READER", None)
    if custom_reader:
        return custom_reader(raw_person)

    if not raw_person["merge_fields"]["ADDRESS"]:
        return ""
    return raw_person["merge_fields"]["ADDRESS"]["city"]


def address(raw_person):
    custom_reader = getattr(settings, "MAILCHIMP_ADDRESS_READER", None)
    if custom_reader:
        return custom_reader(raw_person)

    if not raw_person["merge_fields"]["ADDRESS"]:
        return ""

    addr1 = raw_person["merge_fields"]["ADDRESS"]["addr1"]
    addr2 = raw_person["merge_fields"]["ADDRESS"]["addr2"]
    return addr1 + ("\naddr2" if addr2 else "")


def postcode(raw_person):
    custom_reader = getattr(settings, "MAILCHIMP_ZIP_READER", None)
    if custom_reader:
        return custom_reader(raw_person)

    if not raw_person["merge_fields"]["ADDRESS"]:
        return ""
    return raw_person["merge_fields"]["ADDRESS"]["zip"]


def country(raw_person):
    custom_reader = getattr(settings, "MAILCHIMP_COUNTRY_READER", None)
    if custom_reader:
        return custom_reader(raw_person)

    if not raw_person["merge_fields"]["ADDRESS"]:
        return getattr(settings, "MAILCHIMP_DEFAULT_COUNTRY", "CH")

    return raw_person["merge_fields"]["ADDRESS"]["country"]


def status(raw_person) -> str:
    return raw_person["status"]


def interests_ids(raw_person) -> List[str]:
    interest_indicators = raw_person.get("interests")
    if interest_indicators is None:
        return []
    interests_ids = [
        interest_id
        for interest_id, interested in interest_indicators.items()
        if interested
    ]
    return interests_ids


def email(raw_person) -> str:
    return raw_person["email_address"]


def display_name(raw_person) -> str:
    return (
        f"{raw_person['merge_fields']['FNAME']} {raw_person['merge_fields']['LNAME']}"
    )


def last_name(raw_person) -> str:
    return raw_person["merge_fields"]["LNAME"]


def first_name(raw_person) -> str:
    return raw_person["merge_fields"]["FNAME"]


def language(raw_person) -> str:
    custom_reader = getattr(settings, "MAILCHIMP_LANGUAGE_READER", None)
    if custom_reader:
        return custom_reader(raw_person)
    return raw_person["language"]


def persontype(raw_person):
    custom_reader = getattr(settings, "MAILCHIMP_PERSONTYPE_READER", None)
    if custom_reader:
        return custom_reader(raw_person)
    else:
        return ""



# Generated by Django 3.2.5 on 2021-10-30 02:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contributions", "0003_alter_contribution__import"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contribution",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="contributionimport",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



# flake8: noqa
"""
For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

import os

from basxbread.settings.required import *

# the above will import a set of predefined settings to ensure required
# settings are defined correctly and to reduce verbosity in this file

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "test"  # nosec # can ignore security check for testing key

ALLOWED_HOSTS = ["*"]

# BASXBREAD_DEPENDENCIES are imported in the start import at the top
INSTALLED_APPS = [
    "basxconnect.core",
    "basxconnect.contributions.apps.ContributionsConfig",
] + BASXBREAD_DEPENDENCIES

TEMPLATES[0]["OPTIONS"]["context_processors"].append(
    "basxconnect.core.context_processors.basxconnect_core"
)

ROOT_URLCONF = "basxconnect.core.tests.urls"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}
HAYSTACK_CONNECTIONS = {
    "default": {
        "ENGINE": "haystack.backends.simple_backend.SimpleEngine",
    }
}
HAYSTACK_WHOOSH_STORAGE = "ram"
STATIC_URL = "static/"
STATIC_ROOT = "static"



from basxbread import layout as layout
from basxbread import menu, utils
from basxbread.layout import DEVMODE_KEY
from django.utils.translation import gettext_lazy as _

from .. import models


# New menu group with developer mode checking
class DevGroup(menu.Group):
    def has_permission(self, context):
        return super().has_permission(context) and context["request"].session.get(
            DEVMODE_KEY, False
        )


class SuperUserItem(menu.Item):
    def has_permission(self, context):
        return super().has_permission(context) and context["request"].user.is_superuser


R = layout.grid.Row
C = layout.grid.Col
F = layout.forms.FormField

# MENU ENTRIES ---------------------------------------------------------------------

persongroup = menu.Group(_("Persons"), iconname="group")

menu.registeritem(
    menu.Item(
        utils.Link(
            utils.reverse_model(models.Person, "browse"),
            _("Persons"),
            iconname="group",
            permissions=[utils.permissionname(models.Person, "view")],
        ),
        persongroup,
    )
)

menu.registeritem(
    menu.Item(
        utils.Link(
            utils.reverse_model(models.Vocabulary, "browse"),
            _("Taxonomy"),
            permissions=[utils.permissionname(models.Vocabulary, "view")],
        ),
        menu.settingsgroup,
    )
)

menu.registeritem(
    menu.Item(
        utils.Link(
            utils.reverse(
                "basxconnect.core.views.settings_views.relationshipssettings"
            ),
            _("Relationships"),
            permissions=[utils.permissionname(models.RelationshipType, "view")],
        ),
        menu.settingsgroup,
    )
)



from django.db import models
from django.utils.translation import gettext_lazy as _

from basxconnect.core.models import Person


class Project(models.Model):
    client = models.ForeignKey(
        Person,
        on_delete=models.CASCADE,
        related_name="projects",
        verbose_name=_("Client"),
    )
    name = models.CharField(_("Project name"), max_length=255)
    description = models.TextField(_("Description"), blank=True)

    def __str__(self):
        return f"{self.name} ({self.client})"

    class Meta:
        verbose_name = _("Project")
        verbose_name_plural = _("Projects")



# Generated by Django 4.1.7 on 2023-04-03 05:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("core", "0035_alter_email_type_alter_fax_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Project",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, verbose_name="Project name")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="Description"),
                ),
                (
                    "client",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="projects",
                        to="core.person",
                    ),
                ),
            ],
            options={
                "verbose_name": "Project",
                "verbose_name_plural": "Projects",
            },
        ),
    ]



from django.utils.translation import gettext_lazy as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    BooleanPreference,
    LongStringPreference,
    StringPreference,
)

mailchimp_integration = Section("mailchimp", _("Mailchimp"))


@global_preferences_registry.register
class MailchimpServer(StringPreference):
    section = mailchimp_integration
    name = "server"
    default = ""
    verbose_name = _("Mailchimp server")


@global_preferences_registry.register
class MailchimpListId(StringPreference):
    section = mailchimp_integration
    name = "list_id"
    default = ""
    verbose_name = _("Mailchimp list ID")


@global_preferences_registry.register
class MailchimpSegmentId(StringPreference):
    section = mailchimp_integration
    name = "segment_id"
    default = ""
    verbose_name = _("Mailchimp segment ID")


@global_preferences_registry.register
class MailchimpInterestsCategoryId(StringPreference):
    section = mailchimp_integration
    name = "interests_category_id"
    default = ""
    verbose_name = _("Mailchimp interests category ID")


@global_preferences_registry.register
class MailchimpTag(StringPreference):
    section = mailchimp_integration
    name = "tag"
    default = "Synchronized with Basxconnect"
    verbose_name = _("Mailchimp tag")


@global_preferences_registry.register
class MailchimpBasxconnectTag(StringPreference):
    section = mailchimp_integration
    name = "basxconnect_tag"
    default = "Imported from Mailchimp"
    verbose_name = _("BasxConnect Mailchimp tag")


@global_preferences_registry.register
class MailchimpApiKey(LongStringPreference):
    section = mailchimp_integration
    name = "api_key"
    default = ""
    verbose_name = _("Mailchimp API key")


@global_preferences_registry.register
class MailchimpDisableInterestsKey(BooleanPreference):
    section = mailchimp_integration
    name = "disable_interests"
    default = False
    verbose_name = _("Disable mailing interests")


@global_preferences_registry.register
class MailchimpSynchronizeLanguage(BooleanPreference):
    section = mailchimp_integration
    name = "synchronize_language"
    default = False
    verbose_name = _("Synchronize language with Mailchimp")


@global_preferences_registry.register
class MailchimpAutomaticallySubscribeNewPersons(BooleanPreference):
    section = mailchimp_integration
    name = "automatically_subscribe_new_persons"
    default = False
    verbose_name = _("Automatically subscribe new persons")



# Generated by Django 4.0.6 on 2022-08-19 06:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0031_alter_email_person_alter_email_type_alter_fax_person_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicallegalperson",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Legal Person",
                "verbose_name_plural": "historical Legal Persons",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalnaturalperson",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Natural Person",
                "verbose_name_plural": "historical Natural Persons",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalperson",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Person",
                "verbose_name_plural": "historical Persons",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalpersonassociation",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Person Association",
                "verbose_name_plural": "historical Person Associations",
            },
        ),
        migrations.AlterField(
            model_name="historicallegalperson",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalnaturalperson",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalperson",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalpersonassociation",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



# Generated by Django 3.2.5 on 2021-11-24 08:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0023_auto_20211030_0955"),
    ]

    operations = [
        migrations.AlterField(
            model_name="term",
            name="slug",
            field=models.CharField(blank=True, max_length=255, verbose_name="Slug"),
        ),
    ]



# Generated by Django 3.1.5 on 2021-03-11 17:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0006_auto_20210312_0016"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicallegalperson",
            name="_type",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="core.term",
            ),
        ),
        migrations.AddField(
            model_name="historicalnaturalperson",
            name="_type",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="core.term",
            ),
        ),
        migrations.AddField(
            model_name="historicalperson",
            name="_type",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="core.term",
            ),
        ),
        migrations.AddField(
            model_name="historicalpersonassociation",
            name="_type",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="core.term",
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="core.term",
            ),
        ),
    ]



# Generated by Django 3.2.5 on 2021-09-28 10:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0016_auto_20210928_1534"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="person",
            name="categories",
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from .persons import Person


class RelationshipType(models.Model):
    name = models.CharField(
        _("Name for relationship type"),
        max_length=255,
        help_text="e.g. Person A is member of person B",
    )
    reverse_name = models.CharField(
        _("Backward name for relationship type"),
        max_length=255,
        blank=True,
        help_text="e.g. Person B has person A as a member",
    )

    def save(self, *args, **kwargs):
        if not self.reverse_name:
            self.reverse_name = self.name
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ["name"]
        verbose_name = _("Relationship type")
        verbose_name_plural = _("Relationship types")


class Relationship(models.Model):
    type = models.ForeignKey(RelationshipType, on_delete=models.PROTECT)
    type.verbose_name = _("Type")
    person_a = models.ForeignKey(
        Person, on_delete=models.CASCADE, related_name="relationships_to"
    )
    person_a.verbose_name = _("Person A")
    person_b = models.ForeignKey(
        Person,
        on_delete=models.CASCADE,
        related_name="relationships_from",
    )
    person_b.verbose_name = _("Person B")
    start_date = models.DateField(_("Starts on"), blank=True, null=True)
    end_date = models.DateField(_("Ends on"), blank=True, null=True)

    def __str__(self):
        return "%s is %s of %s" % (self.person_a, self.type, self.person_b)

    class Meta:
        verbose_name = _("Relationship")
        verbose_name_plural = _("Relationships")



import decimal
import re

import htmlgenerator as hg
from basxbread.contrib.document_templates.models import DocumentTemplate
from django.forms import ValidationError
from django.utils.translation import gettext_lazy as _
from djmoney import settings
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    BooleanPreference,
    ChoicePreference,
    DecimalPreference,
    LongStringPreference,
    ModelChoicePreference,
    StringPreference,
)

from basxconnect.core.models import Term

invoicing = Section("invoicing", _("Invoicing"))


@global_preferences_registry.register
class DefaultCurrency(ChoicePreference):
    section = invoicing
    name = "default_currency"
    verbose_name = _("Default currency")
    default = settings.DEFAULT_CURRENCY or "EUR"
    choices = settings.CURRENCY_CHOICES


@global_preferences_registry.register
class UseVAT(BooleanPreference):
    section = invoicing
    name = "default_use_vat"
    verbose_name = _("Use VAT by default")
    default = True


@global_preferences_registry.register
class DefaultVAT(DecimalPreference):
    section = invoicing
    name = "default_vat"
    verbose_name = _("Default VAT")
    default = decimal.Decimal(0)


@global_preferences_registry.register
class UseWHT(BooleanPreference):
    section = invoicing
    name = "default_use_wht"
    verbose_name = _("Use WHT by default")
    default = False


@global_preferences_registry.register
class DefaultWHT(DecimalPreference):
    section = invoicing
    name = "default_wht"
    verbose_name = _("Default WHT")
    default = decimal.Decimal(0)


@global_preferences_registry.register
class AmountsIncludeVAT(BooleanPreference):
    section = invoicing
    name = "amounts_include_vat"
    verbose_name = _("Provided invoicing amounts include VAT by default")
    default = False


@global_preferences_registry.register
class InvoiceNumberGroupingRegex(StringPreference):
    section = invoicing
    name = "invoice_number_grouping_regex"
    verbose_name = _("Invoice number grouping regex")
    help_text = hg.UL(
        "Regex with exactly one group that extracts a grouping value for invoices"
    )
    default = r"(.*)-.*"

    def validate(self, value):
        try:
            groups = re.compile(value).groups
            if groups != 1:
                raise ValidationError(
                    "Requires exactly one regex group, found {groups}"
                )
        except Exception as e:
            raise ValidationError(str(e))


@global_preferences_registry.register
class InvoiceNumberExtractionRegex(StringPreference):
    section = invoicing
    name = "invoice_number_extraction_regex"
    verbose_name = _("Invoice number extraction regex")
    help_text = hg.UL("Regex with exactly one group that extracts a integer value")
    default = r".*-(\d*)"

    def validate(self, value):
        try:
            groups = re.compile(value).groups
            if groups != 1:
                raise ValidationError(
                    "Requires exactly one regex group, found {groups}"
                )
        except Exception as e:
            raise ValidationError(str(e))


@global_preferences_registry.register
class InvoiceNumberTemplate(StringPreference):
    section = invoicing
    name = "invoice_number_template"
    verbose_name = _("Invoice number template")
    help_text = hg.UL(
        hg.LI("Jinja template to generate invoice number"),
        hg.LI("Context variable 'date': current date"),
        hg.LI(
            "Context variable 'counter_next': Next counter number (depending on counter extraction regex)"
        ),
    )
    default = "{{ now()|date('Y') }}-{{ counter_next }}"


@global_preferences_registry.register
class DefaultInvoiceTemplate(ModelChoicePreference):
    section = invoicing
    name = "default_invoice_template"
    verbose_name = _("Default invoice template")
    queryset = DocumentTemplate.objects.filter(
        model__app_label="basxconnect_invoicing", model__model="invoice"
    )
    default = None


@global_preferences_registry.register
class DefaultReceiptTemplate(ModelChoicePreference):
    section = invoicing
    name = "default_receipt_template"
    verbose_name = _("Default receipt template")
    queryset = DocumentTemplate.objects.filter(
        model__app_label="basxconnect_invoicing", model__model="invoice"
    )
    default = None


@global_preferences_registry.register
class InvoiceEmailSubject(StringPreference):
    section = invoicing
    name = "invoice_message_subject_template"
    verbose_name = _("Invoice email subject")
    help_text = hg.UL(
        hg.LI("Jinja template to generate invoice email subject"),
        hg.LI("Context variable 'invoice'"),
        hg.LI("Context variable 'our_company_name'"),
    )
    default = "Invoice {{ invoice.number }} from {{ our_company_name }}"


@global_preferences_registry.register
class InvoiceEmailBody(LongStringPreference):
    section = invoicing
    name = "invoice_message_body_template"
    verbose_name = _("Invoice email body")
    help_text = hg.UL(
        hg.LI("Jinja template to generate invoice email body"),
        hg.LI("Context variable 'invoice'"),
        hg.LI("Context variable 'our_company_name'"),
    )
    default = """Hi {{ invoice.client }}

This is the latest invoice for our services, please find it attached to this message.

Have a nice day!"""


@global_preferences_registry.register
class DefaultInvoiceEmail(ModelChoicePreference):
    section = invoicing
    name = "default_invoice_email_type"
    queryset = Term.objects.filter(vocabulary__slug="emailtype")
    verbose_name = _("Default invoice email")
    default = None


@global_preferences_registry.register
class ReceiptEmailSubject(StringPreference):
    section = invoicing
    name = "receipt_message_subject_template"
    verbose_name = _("Receipt email subject")
    help_text = hg.UL(
        hg.LI("Jinja template to generate receipt email subject"),
        hg.LI("Context variable 'invoice'"),
        hg.LI("Context variable 'our_company_name'"),
    )
    default = "Receipt for invoice {{ invoice.number }} from {{ our_company_name }}"


@global_preferences_registry.register
class ReceiptEmailBody(LongStringPreference):
    section = invoicing
    name = "receipt_message_body_template"
    verbose_name = _("Receipt email body")
    help_text = hg.UL(
        hg.LI("Jinja template to generate receipt email body"),
        hg.LI("Context variable 'invoice'"),
        hg.LI("Context variable 'our_company_name'"),
    )
    default = """Hi {{ invoice.client }}

This is the receipt for invoice {{ invoice.number }}, please find it in the attachment.

Have a nice day!"""



import basxbread
import htmlgenerator as hg
from basxbread import layout as layout
from basxbread.layout.components.button import Button
from basxbread.layout.components.modal import Modal, modal_with_trigger
from basxbread.utils import aslayout, reverse_model
from basxbread.views import AddView
from django import forms
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy
from htmlgenerator import mark_safe

from basxconnect.core import models
from basxconnect.core.models import Term, Vocabulary


class AddTagView(AddView):
    model = models.Term
    fields = ["term", "vocabulary"]

    def form_valid(self, *args, **kwargs):
        ret = super().form_valid(*args, **kwargs)
        ret["HX-Redirect"] = f"{self.request.GET['next']}&new-tag={self.object.id}"
        return ret


@aslayout
def bulk_tag_operation_view(request):
    operation = request.GET["operation"]
    initial = request.GET.get("new-tag")
    if operation not in ["add", "remove"]:
        return HttpResponseBadRequest("invalid GET parameter 'operation'")

    class PersonList(forms.Form):
        persons = forms.ModelMultipleChoiceField(queryset=models.Person.objects.all())

    personlist = PersonList(request.GET)
    if personlist.is_valid():
        persons = [person.pk for person in personlist.cleaned_data["persons"]]
    else:
        return HttpResponseBadRequest("invalid GET parameter 'persons'")

    class BulkTagOperationForm(forms.Form):
        tag = forms.ModelChoiceField(
            queryset=models.Term.objects.filter(vocabulary__slug="tag"),
            required=True,
            initial=initial,
        )

    if request.method == "POST":
        form = BulkTagOperationForm(request.POST)
        if form.is_valid():
            tag = form.cleaned_data.get("tag")
            for person in models.Person.objects.filter(pk__in=persons):
                if operation == "add":
                    person.tags.add(tag)
                else:
                    person.tags.remove(tag)
                person.save()
            return HttpResponseRedirect(reverse_model(models.Person, "browse"))

    form = BulkTagOperationForm()
    count = len(persons)
    if operation == "add":
        header = ngettext_lazy(
            "Add tag to %(count)d person",
            "Add tag to %(count)d persons",
            count,
        ) % {"count": count}
    else:
        header = ngettext_lazy(
            "Remove tag from %(count)d person",
            "Remove tag from %(count)d persons",
            count,
        ) % {"count": count}
    tags_vocabulary_id = (
        getattr(Vocabulary.objects.filter(slug="tag").first(), "id", "") or ""
    )
    return basxbread.layout.forms.Form(
        form,
        hg.H3(header),
        hg.DIV(
            hg.DIV(basxbread.layout.forms.FormField("tag")),
            hg.If(
                operation == "add",
                hg.DIV(
                    modal_with_trigger(
                        Modal.with_ajax_content(
                            heading=_("Create new tag"),
                            url=reverse_model(
                                Term,
                                "ajax_add",
                                query={
                                    "vocabulary": tags_vocabulary_id,
                                    "asajax": True,
                                    "next": mark_safe(
                                        reverse_model(
                                            models.Person,
                                            "bulk-tag-operation",
                                            query={
                                                "operation": operation,
                                                "persons": persons,
                                            },
                                        )
                                    ),
                                },
                            ),
                            submitlabel=_("Save"),
                        ),
                        Button,
                        _("Create new tag"),
                        buttontype="ghost",
                        style="margin-bottom: 2rem;",
                        icon="add",
                    ),
                ),
            ),
            style="display:flex;align-items:end;",
        ),
        layout.forms.helpers.Submit(_("Submit")),
    )



# Generated by Django 4.1.7 on 2023-04-05 04:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("basxconnect_invoicing", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="note",
            field=models.TextField(
                blank=True,
                help_text="Can be displayed on invoice and/or receipt",
                verbose_name="Note to client",
            ),
        ),
    ]



import htmlgenerator as hg
from basxbread import layout
from basxbread.layout import ObjectFieldLabel
from basxbread.layout.components.icon import Icon
from django.utils.translation import gettext_lazy as _

from basxconnect.core import models
from basxconnect.core.layouts.editperson.common import addresses, base_data, utils
from basxconnect.core.layouts.editperson.common.utils import (
    display_label_and_value,
    person_metadata,
)

R = layout.grid.Row
C = layout.grid.Col


def base_data_tab():
    return layout.tabs.Tab(
        _("Personal data"),
        utils.grid_inside_tab(
            R(
                personal_data(),
                person_metadata(models.NaturalPerson),
            ),
            base_data.common_tiles(),
        ),
    )


def personal_data():
    displayed_fields = (
        [
            utils.display_field_label_and_value(field)
            for field in [
                "salutation",
                "title",
                "name",
                "first_name",
                "last_name",
            ]
        ]
        + [
            hg.If(
                hg.C("object.middle_name"),
                utils.display_field_label_and_value("middle_name"),
            )
        ]
        + [
            hg.If(
                hg.C("object.maiden_name"),
                utils.display_field_label_and_value("maiden_name"),
            )
        ]
        + [
            utils.display_field_label_and_value(field)
            for field in [
                "gender",
                "date_of_birth",
                "place_of_birth",
                "profession",
            ]
        ]
        + [
            hg.If(
                hg.C("object.deceased"),
                display_label_and_value(
                    ObjectFieldLabel("deceased"),
                    hg.If(hg.C("object.deceased"), _("Yes"), _("No")),
                ),
            ),
            hg.If(
                hg.C("object.deceased"),
                utils.display_field_label_and_value("decease_date"),
            ),
        ]
    )
    return utils.tile_col_edit_modal_displayed_fields(
        _("Personal Data"),
        models.NaturalPerson,
        "ajax_edit_personal_data",
        Icon("user--profile"),
        displayed_fields,
    )


def contact_details_naturalperson(request):
    return hg.BaseElement(
        R(
            addresses.postals(),
            addresses.numbers(request),
        ),
        R(
            addresses.email(request),
            addresses.urls(request),
        ),
        R(
            base_data.tags(),
            base_data.other(),
        ),
    )



# Generated by Django 3.2.5 on 2021-10-18 09:44

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0021_auto_20211007_0932"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="email",
            name="status",
        ),
        migrations.RemoveField(
            model_name="fax",
            name="status",
        ),
        migrations.RemoveField(
            model_name="phone",
            name="status",
        ),
        migrations.RemoveField(
            model_name="postal",
            name="status",
        ),
        migrations.RemoveField(
            model_name="web",
            name="status",
        ),
    ]



# Generated by Django 3.2.13 on 2022-06-02 07:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0029_auto_20220429_1138"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalnaturalperson",
            name="place_of_birth",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="Place of Birth"
            ),
        ),
        migrations.AddField(
            model_name="naturalperson",
            name="place_of_birth",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="Place of Birth"
            ),
        ),
    ]



import random

from basxbread import layout
from basxbread.contrib.languages.fields import LanguageField
from basxbread.utils import get_concrete_instance
from basxbread.utils.inheritancemanager import InheritanceManager
from django import forms
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from simple_history.models import HistoricalRecords

from .utils import Note, Term


def preferred_languages_choices(field, request, instance):
    from .. import settings

    return settings.PREFERRED_LANGUAGES


class PersonManager(InheritanceManager):
    def get_queryset(self):
        return super().get_queryset().select_subclasses()

    def trash(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(deleted=True)

    def not_deleted(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(deleted=False)


# don't delete this function, other initial migrations for new projects will fail
def random_personid():
    # this is not security related, we can disable checks here
    return f"__placeholder__{random.randint(100000, 999999)}"  # nosec


class Person(models.Model):
    deleted = models.BooleanField(
        _("Deleted"), blank=True, default=False, editable=False
    )
    personnumber = models.CharField(
        _("Person number"),
        max_length=255,
        unique=True,
        default=random_personid,
    )
    personnumber.sorting_name = "personnumber__int"
    name = models.CharField(_("Display name"), max_length=255, blank=True)
    # this one is to solve a problem when sorting names from
    # naturalperson and other types of person
    default_sorting_name = models.CharField(
        _("Default Sorting Name"), max_length=255, blank=True, editable=False
    )
    active = models.BooleanField(_("Active"), default=True, editable=False)
    salutation_letter = models.CharField(
        _("Salutation Letter"),
        max_length=255,
        blank=True,
    )
    preferred_language = LanguageField(_("Preferred Language"), blank=True)
    preferred_language.lazy_choices = preferred_languages_choices

    tags = models.ManyToManyField(
        Term,
        blank=True,
        limit_choices_to={"vocabulary__slug": "tag"},
        # db_table="core_person_categories",
    )
    tags.verbose_name = _("Tags")

    primary_postal_address = models.ForeignKey(
        "core.Postal",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="primary_address_for",
    )
    primary_postal_address.verbose_name = _("Primary postal address")
    primary_postal_address.lazy_choices = (
        lambda field, request, instance: instance.core_postal_list.all()
        if hasattr(instance, "core_postal_list")
        else None
    )
    primary_email_address = models.ForeignKey(
        "core.Email",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="primary_email_for",
    )
    primary_email_address.verbose_name = _("Primary email address")
    primary_email_address.lazy_choices = (
        lambda field, request, instance: instance.core_email_list.all()
        if hasattr(instance, "core_email_list")
        else None
    )
    primary_phonenumber = models.ForeignKey(
        "core.Phone",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="primary_phone_for",
    )
    primary_phonenumber.verbose_name = _("Primary phonenumber")
    primary_phonenumber.lazy_choices = (
        lambda field, request, instance: instance.core_phone_list.all()
        if hasattr(instance, "core_phone_list")
        else None
    )

    remarks = models.TextField(_("Remarks"), blank=True)
    notes = GenericRelation(Note)
    history = HistoricalRecords(inherit=True)
    _maintype = models.CharField(  # only used internally for filtering or sorting
        max_length=32,
        choices=(
            ("person", "person"),
            ("naturalperson", "naturalperson"),
            ("legalperson", "legalperson"),
            ("personassociation", "personassociation"),
        ),
        editable=False,
    )

    _type = models.ForeignKey(
        Term, on_delete=models.SET_NULL, null=True, blank=True, related_name="+"
    )
    _type.verbose_name = _("Person Type")

    objects = PersonManager()

    def __str__(self):
        return self.name

    def maintype(self):
        return get_concrete_instance(self)._meta.verbose_name

    maintype.verbose_name = _("Main Type")
    maintype.sorting_name = "_maintype"

    def status(self):
        return (
            layout.tag.Tag(_("active"), tag_color="green")
            if self.active
            else layout.tag.Tag(_("inactive"), tag_color="red")
        )

    status.verbose_name = _("Status")
    status.sorting_name = "active"

    def save(self, *args, **kwargs):
        created = self.pk is None
        # automatically generate a person number if empty
        # if this object is just beeing create we do not have
        # a person number (because there is no pk) and we need
        # to do the same thing again after a call to super.save
        # and then save again
        # the random-things is a bit ugly but if we want to enforce uniqueness
        # on the database level we have to use this one during creation

        if not self._maintype:
            self._maintype = "person"
        if self.pk is not None:
            if hasattr(self, "core_postal_list"):
                active_postals = [
                    postal
                    for postal in self.core_postal_list.all()
                    if postal.valid_until is None
                    or postal.valid_until >= timezone.now().date()
                ]
                if len(active_postals) == 0:
                    self.primary_postal_address = None
                elif (
                    len(active_postals) == 1
                    or self.primary_postal_address is None
                    or (
                        self.primary_postal_address.valid_until
                        and self.primary_postal_address.valid_until
                        < timezone.now().date()
                    )
                ):
                    self.primary_postal_address = active_postals[0]

            else:
                self.primary_postal_address = None

            if hasattr(self, "core_email_list"):
                if self.core_email_list.all().count() == 1 or (
                    self.core_email_list.all().count() > 0
                    and self.primary_email_address is None
                ):
                    self.primary_email_address = self.core_email_list.first()
            else:
                self.primary_email_address = None

            if hasattr(self, "core_phone_list"):
                if (
                    self.core_phone_list.all().count() == 1
                    or self.primary_phonenumber is None
                ):
                    self.primary_phonenumber = self.core_phone_list.first()
            else:
                self.primary_phonenumber = None

        super().save(*args, **kwargs)
        if self.personnumber.startswith("__placeholder__"):
            self.personnumber = str(self.pk)
            super().save(update_fields=["personnumber"])

        self.default_sorting_name = self.name

        # this signal needs to be sent manually in order to trigger the search-index update
        # Django does only send a signal for the child-model but our search-index only observes
        # this base model. It is only needed when creating for some reason...
        models.signals.post_save.send(
            sender=Person,
            instance=getattr(self, "person_ptr", self),
            created=created,
            update_fields=kwargs.get("update_fields"),
            raw=False,
            using=kwargs.get("using"),
        )

    def search_index_snippet(self):
        concrete = get_concrete_instance(self)
        if concrete != self and hasattr(concrete, "search_index_snippet"):
            return concrete.search_index_snippet()
        addr = self.primary_postal_address
        pieces = [self.name]
        if addr:
            pieces.append(addr.address.replace("\n", " "))
            pieces.append(f"{addr.postcode} {addr.city}")
            pieces.append(addr.country.code)

        return ", ".join(pieces)

    class Meta:
        ordering = ["name"]
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")


class NaturalPerson(Person):
    first_name = models.CharField(_("First Name"), max_length=255, blank=True)
    middle_name = models.CharField(_("Middle Name"), max_length=255, blank=True)
    last_name = models.CharField(_("Last Name"), max_length=255, blank=True)
    maiden_name = models.CharField(_("Maiden Name"), max_length=255, blank=True)

    autogenerate_displayname = models.BooleanField(
        _("Autogenerate Display Name"), default=True
    )
    title = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "title"},
        related_name="title_persons",
        verbose_name=_("Title"),
    )
    title.formfield_kwargs = {"widget": forms.RadioSelect()}

    salutation = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "salutation"},
        related_name="+",
        verbose_name=_("Salutation"),
    )
    salutation.formfield_kwargs = {"widget": forms.RadioSelect()}

    form_of_address = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "form_of_address"},
        related_name="+",
        verbose_name=_("Form of address"),
    )
    form_of_address.formfield_kwargs = {"widget": forms.RadioSelect()}

    profession = models.CharField(
        _("Profession"),
        max_length=255,
        blank=True,
    )
    date_of_birth = models.DateField(_("Date of Birth"), blank=True, null=True)
    place_of_birth = models.CharField(_("Place of Birth"), max_length=255, blank=True)
    deceased = models.BooleanField(_("Deceased"), default=False)
    decease_date = models.DateField(_("Deceased Date"), blank=True, null=True)
    gender = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "gender"},
        related_name="gender_persons",
        verbose_name=_("Gender"),
    )
    gender.formfield_kwargs = {"widget": forms.RadioSelect()}

    type = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "naturaltype"},
    )
    type.verbose_name = _("Person Type")

    def age(self):
        if not self.date_of_birth:
            return None
        today = timezone.now().date()
        birth = self.date_of_birth
        return (
            today.year
            - birth.year
            - ((today.month, today.day) < (birth.month, birth.day))
        )

    age.verbose_name = _("Age")

    def save(self, *args, **kwargs):
        if self.autogenerate_displayname:
            self.name = self.first_name + " " + self.last_name
        self.default_sorting_name = self.last_name
        if self.decease_date:
            self.deceased = True
        self._type = (
            self.type
            or global_preferences_registry.manager()["persons__default_naturaltype"]
        )
        self._maintype = "naturalperson"
        super().save(*args, **kwargs)

    class Meta:
        ordering = ["last_name", "first_name"]
        verbose_name = _("Natural Person")
        verbose_name_plural = _("Natural Persons")


class LegalPerson(Person):
    name_addition = models.CharField(_("Addition name"), max_length=255, blank=True)
    type = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "legaltype"},
    )
    type.verbose_name = _("Type")

    def save(self, *args, **kwargs):
        self._maintype = "legalperson"
        self._type = (
            self.type
            or global_preferences_registry.manager()["persons__default_legaltype"]
        )
        super().save(*args, **kwargs)

    class Meta:
        ordering = ["name"]
        verbose_name = _("Legal Person")
        verbose_name_plural = _("Legal Persons")


class PersonAssociation(Person):
    type = models.ForeignKey(
        Term,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        limit_choices_to={"vocabulary__slug": "associationtype"},
    )
    type.verbose_name = _("Type")

    def save(self, *args, **kwargs):
        self._maintype = "personassociation"
        self._type = (
            self.type
            or global_preferences_registry.manager()["persons__default_associationtype"]
        )
        super().save(*args, **kwargs)

    class Meta:
        ordering = ["name"]
        verbose_name = _("Person Association")
        verbose_name_plural = _("Person Associations")



# Generated by Django 3.2.12 on 2022-04-04 03:58

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0027_auto_20211223_1948"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="term",
            options={"verbose_name": "Term", "verbose_name_plural": "Terms"},
        ),
        migrations.AlterOrderWithRespectTo(
            name="term",
            order_with_respect_to="vocabulary",
        ),
    ]



# Generated by Django 3.2.4 on 2021-11-22 13:30

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mailer_integration", "0017_auto_20211122_1219"),
    ]

    operations = [
        migrations.AlterField(
            model_name="synchronizationperson",
            name="sync_status",
            field=models.CharField(
                blank=True,
                choices=[
                    ("new", "Newly added to BasxConnect"),
                    ("import_error", "Not added to BasxConnect"),
                    ("synced_previously", "Synchronized previously but not this time"),
                ],
                max_length=255,
                verbose_name="Synchronization Status",
            ),
        ),
    ]



# Generated by Django 3.2.4 on 2021-08-18 07:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("core", "0011_auto_20210726_2047"),
    ]

    operations = [
        migrations.CreateModel(
            name="Interest",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="MailingPreferences",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("status", models.CharField(max_length=50)),
                ("interests", models.ManyToManyField(to="mailer_integration.Interest")),
                (
                    "person",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, to="core.person"
                    ),
                ),
            ],
        ),
    ]



import datetime

from django.conf import settings as djangosettings
from django.db import models
from django.utils.translation import gettext_lazy as _
from djmoney.contrib.exchange.models import convert_money
from djmoney.models.fields import CurrencyField
from djmoney.settings import CURRENCY_CHOICES
from dynamic_preferences.registries import global_preferences_registry
from moneyed import Money

from basxconnect.core.models import Person

from . import settings


class ContributionImport(models.Model):
    date = models.DateTimeField(_("Date"), default=datetime.datetime.now)
    importfile = models.FileField(_("Importfile"))
    user = models.ForeignKey(
        djangosettings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )
    user.verbose_name = _("User")

    def bookingrange(self):
        return f"{self.contributions.last().date} - {self.contributions.first().date}"

    bookingrange.verbose_name = _("Booking range")

    def numberofbookings(self):
        return self.contributions.count()

    numberofbookings.verbose_name = _("Number of bookings")

    def totalamount(self):
        global_preferences = global_preferences_registry.manager()
        return sum(
            [
                convert_money(
                    c.amount_formatted(), global_preferences["contributions__currency"]
                )
                for c in self.contributions.all()
            ]
        )

    totalamount.verbose_name = _("Total amount")

    def __str__(self):
        return f"{self.date.date()}"

    class Meta:
        verbose_name = _("Contribution Import")
        verbose_name_plural = _("Contributions Imports")
        ordering = ["-date"]


class Contribution(models.Model):
    _import = models.ForeignKey(
        ContributionImport,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="contributions",
    )
    _import.verbose_name = _("Import")
    person = models.ForeignKey(
        Person, null=True, on_delete=models.SET_NULL, related_name="contributions"
    )
    person.verbose_name = _("Person")

    date = models.DateField(_("Date"))
    note = models.CharField(_("Note"), max_length=255)
    debitaccount = models.CharField(_("Debit Account"), max_length=32)
    creditaccount = models.CharField(_("Credit Account"), max_length=32)
    amount = models.DecimalField(_("Amount"), max_digits=12, decimal_places=2)
    currency = CurrencyField(_("Currency"), choices=CURRENCY_CHOICES)
    currency.lazy_choices = lambda *args: settings.PREFERRED_CURRENCIES

    def amount_formatted(self):
        return Money(self.amount, self.currency)

    amount_formatted.verbose_name = _("Amount")

    def __str__(self):
        return f"{self.date}: {self.person} {self.amount_formatted()}"

    class Meta:
        verbose_name = _("Contribution")
        verbose_name_plural = _("Contributions")
        ordering = ["-date"]



from basxbread.utils.urls import autopath, default_model_paths, model_urlname

import basxconnect.mailer_integration.views
from basxconnect.mailer_integration import models

urlpatterns = [
    autopath(basxconnect.mailer_integration.views.mailer_synchronization_view),
    autopath(
        basxconnect.mailer_integration.views.EditSubscriptionView.as_view(
            model=models.Subscription,
        ),
        urlname=model_urlname(models.Subscription, "ajax_edit"),
    ),
    autopath(
        basxconnect.mailer_integration.views.AddSubscriptionView.as_view(
            model=models.Subscription,
        ),
        urlname=model_urlname(models.Subscription, "ajax_add"),
    ),
    *default_model_paths(models.SynchronizationResult),
]



import htmlgenerator as hg
from basxbread import layout
from basxbread.utils import Link, ModelHref, reverse
from basxbread.views import BrowseView
from django.utils.translation import gettext_lazy as _

from basxconnect.core.models import RelationshipType, Term, Vocabulary

R = layout.grid.Row
C = layout.grid.Col
F = layout.forms.FormField


def relationshipssettings(request):
    return layout.grid.Grid(
        R(C(hg.H3(_("Relationships")))),
        R(
            C(
                layout.datatable.DataTable.from_queryset(
                    RelationshipType.objects.all(),
                    columns=["name"],
                    primary_button=layout.button.Button.from_link(
                        Link(
                            href=ModelHref(
                                RelationshipType,
                                "add",
                                query={
                                    "next": reverse(
                                        "basxconnect.core.views.settings_views.relationshipssettings"
                                    )
                                },
                            ),
                            label=_("Add %s") % RelationshipType._meta.verbose_name,
                        ),
                        icon=layout.icon.Icon("add", size=20),
                    ),
                    rowactions=[
                        Link(
                            label=_("Edit"),
                            href=ModelHref(
                                RelationshipType,
                                "edit",
                                kwargs={"pk": hg.C("row.pk")},
                                query={
                                    "next": reverse(
                                        "basxconnect.core.views.settings_views.relationshipssettings"
                                    )
                                },
                            ),
                            iconname="edit",
                        ),
                        Link(
                            label=_("Delete"),
                            href=ModelHref(
                                RelationshipType,
                                "delete",
                                kwargs={"pk": hg.C("row.pk")},
                                query={
                                    "next": reverse(
                                        "basxconnect.core.views.settings_views.relationshipssettings"
                                    )
                                },
                            ),
                            iconname="trash-can",
                        ),
                    ],
                    backurl=reverse(
                        "basxconnect.core.views.settings_views.relationshipssettings"
                    ),
                ),
            )
        ),
        gutter=False,
    )


def generate_term_datatable(title, vocabulary_slug):
    """Helper function to display a table for all terms of a certain term, currently always returns to the personsettings view"""

    cat = Vocabulary.objects.filter(slug=vocabulary_slug).first() or ""
    return layout.datatable.DataTable.from_queryset(
        Term.objects.filter(vocabulary__slug=vocabulary_slug),
        columns=["term"],
        title=title,
        primary_button=layout.button.Button.from_link(
            Link(
                href=ModelHref(
                    Term, "add", query={"vocabulary": cat.id}, return_to_current=True
                ),
                label=_("Add %s") % cat,
            ),
            icon=layout.icon.Icon("add", size=20),
        ),
        prevent_automatic_sortingnames=True,
        rowclickaction=BrowseView.gen_rowclickaction("edit", return_to_current=True),
        rowactions=[BrowseView.deletelink()],
    )



from basxbread import layout
from basxbread.layout.components.icon import Icon
from django.utils.translation import gettext_lazy as _

from basxconnect.core import models
from basxconnect.core.layouts.editperson.common import base_data
from basxconnect.core.layouts.editperson.common.utils import (
    grid_inside_tab,
    person_metadata,
    tile_col_edit_modal,
)

R = layout.grid.Row


def base_data_tab():
    return layout.tabs.Tab(
        _("Base data"),
        grid_inside_tab(
            R(
                tile_col_edit_modal(
                    _("Base Data"),
                    models.PersonAssociation,
                    "ajax_edit",
                    Icon("building"),
                    [
                        "name",
                        "preferred_language",
                        "salutation_letter",
                    ],
                ),
                person_metadata(models.PersonAssociation),
            ),
            base_data.common_tiles(),
        ),
    )



# Generated by Django 3.1.7 on 2021-03-23 07:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0007_auto_20210312_0054"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicallegalperson",
            name="deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted"
            ),
        ),
        migrations.AddField(
            model_name="historicalnaturalperson",
            name="deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted"
            ),
        ),
        migrations.AddField(
            model_name="historicalperson",
            name="deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted"
            ),
        ),
        migrations.AddField(
            model_name="historicalpersonassociation",
            name="deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted"
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="deleted",
            field=models.BooleanField(
                blank=True, default=False, verbose_name="Deleted"
            ),
        ),
    ]



import htmlgenerator as hg
from basxbread import layout
from basxbread.views import AddView, EditView
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from basxconnect.core import models
from basxconnect.core.views.person import search_person_view


def search_select_person(person: str):
    return layout.forms.FormField(
        person,
        widgetclass=layout.search_select.SearchSelect,
        backend=layout.search.SearchBackendConfig(
            reverse_lazy(
                "basxconnect.core.views.person.search_person_view.searchperson"
            ),
            result_selector=f".{search_person_view.ITEM_CLASS}",
            result_label_selector=f".{search_person_view.ITEM_LABEL_CLASS}",
            result_value_selector=f".{search_person_view.ITEM_VALUE_CLASS}",
        ),
    )


formfields = [
    layout.forms.FormField("type"),
    search_select_person("person_a"),
    search_select_person("person_b"),
    layout.forms.FormField("start_date"),
    layout.forms.FormField("end_date"),
]


class EditRelationshipView(EditView):
    model = models.Relationship

    def get_layout(self):
        return layout.grid.Grid(
            hg.H3(_("Edit Relationship")),
            layout.grid.Row(
                layout.grid.Col(
                    layout.forms.Form(
                        hg.C("form"), hg.DIV(*formfields), layout.forms.helpers.Submit()
                    ),
                    width=4,
                )
            ),
            gutter=False,
        )


class AddRelationshipView(AddView):
    model = models.Relationship

    def get_layout(self):
        if self.ajax_urlparameter in self.request.GET:
            return layout.forms.Form(hg.C("form"), hg.BaseElement(*formfields))
        else:
            return layout.grid.Grid(
                hg.H3(_("Add Relationship")),
                layout.grid.Row(
                    layout.grid.Col(
                        layout.forms.Form(
                            hg.C("form"),
                            hg.DIV(*formfields),
                        ),
                        width=4,
                    )
                ),
                gutter=False,
            )



from setuptools import find_packages, setup


def readme():
    with open("README.md") as f:
        return f.read()


with open("basxconnect/__init__.py") as f:
    # magic n stuff
    version = (
        [i for i in f.readlines() if "__version__" in i][-1]
        .split("=", 1)[1]
        .strip()
        .strip('"')
    )

setup(
    name="basxconnect",
    version=version,
    description="CRM application for non-profit organizations",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/basxsoftwareassociation/basxconnect",
    author="basx Software Association",
    author_email="sam@basx.dev",
    license="New BSD License",
    install_requires=[
        "basx-bread",
        "django-formtools == 2.3",
        "djangorestframework",
        "chardet",
        "tablib",
        "dateparser",
    ],
    extras_require={
        "testing": ["basx-bread[testing]"],
        "mailer-integration": ["mailchimp_marketing"],
    },
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



# Generated by Django 4.0.5 on 2022-06-02 10:03

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0030_auto_20220602_1434"),
    ]

    operations = [
        migrations.AlterField(
            model_name="email",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_list",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="email",
            name="type",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"vocabulary__slug": "emailtype"},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="type_%(app_label)s_%(class)s_list",
                to="core.term",
            ),
        ),
        migrations.AlterField(
            model_name="fax",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_list",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="phone",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_list",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="postal",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_list",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="postal",
            name="type",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"vocabulary__slug": "addresstype"},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="type_%(app_label)s_%(class)s_list",
                to="core.term",
            ),
        ),
        migrations.AlterField(
            model_name="web",
            name="person",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_list",
                to="core.person",
            ),
        ),
        migrations.AlterField(
            model_name="web",
            name="type",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"vocabulary__slug": "urltype"},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="type_%(app_label)s_%(class)s_list",
                to="core.term",
            ),
        ),
    ]



# Generated by Django 3.2.4 on 2021-11-19 12:42

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mailer_integration", "0012_auto_20211119_0830"),
    ]

    operations = [
        migrations.AddField(
            model_name="mailingpreferences",
            name="latest_sync",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="synchronized_subscription",
                to="mailer_integration.synchronizationresult",
            ),
        ),
    ]



from itertools import chain

from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db.models.fields.related import RelatedField

try:
    from itertools import pairwise
except ImportError:
    # python<3.10 compatability
    from itertools import tee

    def pairwise(iterable):
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)


import htmlgenerator as hg
from basxbread import layout
from django.utils.translation import gettext_lazy as _

from basxconnect.core.layouts.editperson.common import utils

R = layout.grid.Row
C = layout.grid.Col


def history_tab():
    """
    new_record, old_record = p.history.all()
    delta = new_record.diff_against(old_record)
    for change in delta.changes:
        print("{} changed from {} to {}".format(change.field, change.old, change.new))
    """

    def changes(c):
        if c["row"][1] is not None:
            return list(c["row"][0].diff_against(c["row"][1]).changes)
        return ()

    def haschanges(a, b):
        return b is not None and len(a.diff_against(b).changes) > 0

    def historyentries(c):
        return (
            (i, j)
            for i, j in pairwise(chain(c["object"].history.all(), [None]))
            if haschanges(i, j)
        )

    def fieldname(c):
        try:
            return c["object"]._meta.get_field(c["change"].field).verbose_name
        except FieldDoesNotExist:
            return c["change"].field.replace("_", " ").capitalize()

    def newfieldvalue(c):
        try:
            field = c["object"]._meta.get_field(c["change"].field)
            if isinstance(field, RelatedField):
                try:
                    return field.related_model.objects.get(id=int(c["change"].new))
                except field.related_model.DoesNotExist:
                    return hg.SPAN(_("<Value has been deleted>"), style="color: red")
        except FieldDoesNotExist:
            pass
        return c["change"].new

    def oldfieldvalue(c):
        try:
            field = c["object"]._meta.get_field(c["change"].field)
            if isinstance(field, RelatedField):
                ret = field.related_model.objects.filter(
                    id=int(c["change"].old)
                ).first()
                return ret or hg.SPAN(_("<Value has been deleted>"), style="color: red")
        except FieldDoesNotExist:
            pass
        return c["change"].old

    return layout.tabs.Tab(
        _("History"),
        utils.grid_inside_tab(
            R(
                utils.tiling_col(
                    layout.components.datatable.DataTable(
                        row_iterator=hg.F(historyentries),
                        columns=[
                            layout.components.datatable.DataTableColumn(
                                _("Date"),
                                layout.localize(
                                    layout.localtime(hg.C("row")[0].history_date).date()
                                ),
                            ),
                            layout.components.datatable.DataTableColumn(
                                _("Time"),
                                layout.localize(
                                    layout.localtime(hg.C("row")[0].history_date).time()
                                ),
                            ),
                            layout.components.datatable.DataTableColumn(
                                _("User"),
                                hg.C("row")[0].history_user,
                            ),
                            layout.components.datatable.DataTableColumn(
                                _("Changes"),
                                hg.UL(
                                    hg.Iterator(
                                        hg.F(changes),
                                        "change",
                                        hg.LI(
                                            hg.SPAN(
                                                hg.F(fieldname),
                                                style="font-weight: 600",
                                            ),
                                            ": ",
                                            hg.SPAN(
                                                hg.If(
                                                    hg.C("change").old,
                                                    hg.F(oldfieldvalue),
                                                    settings.HTML_NONE,
                                                ),
                                                style="text-decoration: line-through;",
                                            ),
                                            " -> ",
                                            hg.SPAN(
                                                hg.If(
                                                    hg.C("change").new,
                                                    hg.F(newfieldvalue),
                                                    settings.HTML_NONE,
                                                )
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ],
                    ).with_toolbar(_("Changes"))
                )
            ),
        ),
    )
